#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <limits.h>
#include <getopt.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "br_cmd.h"

#define VERSION "0.03a"

/*

br (BottleRocket)

Control software for the X10(R) FireCracker wireless computer
interface kit.

(c) 1999 Tymm Twillman (tymm@acm.org).  Free Software.  LGPL applies.
  No warranties expressed or implied.

Have fun with it and if you do anything really cool, send an email and let me
know.


*/


/*
 * NetBSD apparently doesn't use the ttyS stuff... Not sure about FreeBSD, but
 *  /dev/cua works there.  It's being removed from Linux, however...
 */

#ifndef X10_DEVICE
#define X10_DEVICE "/dev/cua0"
#endif

extern char **environ;

char *MyName = "bottlerocket";   /* Default */

void usage()
{
	fprintf(stderr, "BottleRocket version %s                        \n", VERSION);
	fprintf(stderr, "(c) 1999 - Tymm Twillman                       \n");
	fprintf(stderr, "                                               \n");
	fprintf(stderr, "Purpose:  Control X10 \"FireCracker\" wireless \n");
	fprintf(stderr, "           home-control interface things.      \n");
	fprintf(stderr, "                                               \n");
	fprintf(stderr, "Usage: %s [-x x10 device] <A-P> <1-16> (on | off)\n",
      MyName);
 	fprintf(stderr, "       %s [-r <times to repeat>] <A-P> (bright | dim) "
      "\n", MyName); 	
    fprintf(stderr, "                                               \n"); 	
    fprintf(stderr, "Examples: %s A1 on                             \n",
	  MyName);
	fprintf(stderr, "          %s P8 off                            \n",
	  MyName);
	fprintf(stderr, "          %s B dim                             \n",
	  MyName);
	fprintf(stderr, "                                               \n");
	fprintf(stderr, " OR you can use the x10-amh compatability with \n");
	fprintf(stderr, "          %s [-c <A-P>] (-f | -n) <1-16>       \n",
      MyName);
	fprintf(stderr, " where -f turns off a device, -n turns it on.  \n");
	
}

int get_cmd(char *arg)
{
	if (!strcasecmp("on", arg)) {
		return ON;
	} else if (!strcasecmp("off", arg)) {
		return OFF;
	} else if (!strcasecmp("dim", arg)) {
		return DIM;
	} else if (!strcasecmp("bright", arg)) {
		return BRIGHT;
	}

	return -1;
}

int get_device(char *arg, char cmd)
{
	char device;


	/*
	 * Figure out the numeric part of the address of a unit (which device
	 *  in a group) we want to talk to
	 *
	 * If the command is a "dim" command, the device should be set to 0.
	 *  Unfortunately, some of the address lines are used for the dim/
	 *  bright commands.
	 */

	if ((cmd == DIM) || (cmd == BRIGHT)) {
        if (strlen(arg))
	        return -1;
		device = 0;
	} else {

		if ((!strlen(arg)) || (strlen(arg) > 2))
			return -1;
		device = (char)atoi(arg) - 1;
		if ((device < 0) || (device > 15))
			return -1;
	}

	return device;
}

int get_group(char *arg)
{
	/*
	 * Get the letter part of a unit address (what "group" of units 
	 *  requested one is in)
	 */

	int group;


	if (!strlen(arg))
		return -1;

	group = toupper(arg[0]) - 'A';
	if ((group > 15) || (group < 0))
		return -1;

	return group;
}
	
int main(int argc, char *argv[])
{
    char device = 0;
    char group = 0;
    int cmd;
    int fd;
    char *x10_dev = NULL;
    int unit;
    int opt;
    int repeats = 1;
    int amh_compat = 0;
    int x10_dev_set = 0;
    int immutable_device = 0;
    int amh_house_set = 0;
    int amh_dev_set = 0;
    int repeats_set = 0;
    int inverse_set = 0;

/*
 * You could take out these lines if you're gonna run this SUID and want
 *  to let normal users change the x10 device by environment variables/command
 *  line.  You could also do a "chmod -R 0777 /".  The former would probably
 *  be less dangerous, but do you want to take the risk?
 *
 *  Anyhow, it's safer to chmod the device for normal users, as long as you
 *   remember to change it back before you hook up anything really useful.
 */

#ifdef HAS_GETRUID
    if (getruid() != 0)
#else
    if (getuid() != 0)
#endif
        immutable_device = 1;

    /*
     * Who am I?
     */

    if (argc > 0)
	MyName = argv[0];

    if (!immutable_device) {
        if ((environ) && (x10_dev = getenv("X10_PORTNAME"))) {
           if (strlen(x10_dev) > PATH_MAX) {
                fprintf(stderr, "%s:  X10_PORTNAME environment variable "
	          "is too long!\n", MyName);
                exit(22);
           }
        } else {
            x10_dev = X10_DEVICE;
        }
    } else {
        x10_dev = X10_DEVICE;
    }

    while ((opt = getopt(argc, argv, "ir:x:c:f:n:")) != EOF) {
	
        switch(opt) {
            case 'i':
                /*
                 * Undocumented feature!! Figure it out.
                 */
	
                if (inverse_set) {
                    fprintf(stderr, "%s:  You can only specify one -i.\n",
                      MyName);
                    exit(22);
                }
	
                inverse_set = 1;
                break;

            case 'r':
                if (repeats_set) {
                    fprintf(stderr, "%s:  You can only specify one -r.\n",
                      MyName);
                    exit(22);
                }

                repeats = atoi(optarg);
                break;

            case 'x':
                if (immutable_device) {
                    fprintf(stderr, "%s:  You must be root to use a different " 
                      "x10 device.\n", MyName);
                    exit(22);	
                }

                if (x10_dev_set) {
                    fprintf(stderr, "%s:  I can only use one x10 device at "
                      "a time!\n", MyName);
                    exit(22);
                }
	
                x10_dev_set = 1;
	
                if ((x10_dev = optarg) == NULL) {
                    usage();
                    exit(22);
                }
	
                if (strlen(x10_dev) > PATH_MAX) {
                    fprintf(stderr, "%s:  x10 device name (specified with -d "
                      "option) is too long!\n", MyName);
                    exit(22);
                }
	
                break;
	
            case 'c':
                if (amh_house_set) {
                    fprintf(stderr, "%s:  Only one -c can be specified.\n",
                      MyName);
                    exit(22);
                }

                amh_house_set = 1;
                amh_compat = 1;

                if ((group = get_group(optarg)) < 0) {
                    fprintf(stderr, "%s:  Invalid x10 module specified.\n",
                      MyName);
                    exit(22);
                }
	
                break;
	
            case 'f':
                if (amh_dev_set) {
                    fprintf(stderr, "%s:  Only one -f or -n can be "
		      "specified.\n", MyName);
                    exit(22);
                }
	
                amh_dev_set = 1;
                amh_compat = 1;
                cmd = OFF;

                if ((device = get_device(optarg, cmd)) < 0) {
                    fprintf(stderr, "%s:  Invalid x10 module specified.\n",
                      MyName);
                    exit(22);
                }
	
                break;
	
            case 'n':
                if (amh_dev_set) {
                    fprintf(stderr, "%s:  Only one -f or -n can be "
		      "specified.\n", MyName);
                    exit(22);
                }
	
                amh_dev_set = 1;
                amh_compat = 1;
                cmd = ON;

    	        if ((device = get_device(optarg, cmd)) < 0) {
                    fprintf(stderr, "%s:  Invalid x10 module specified.\n",
                      MyName);
                    exit(22);
                }
	
                break;
	
            default:
                usage();
                exit(22);
        }
    }			

    if (amh_compat) {
        if ((argv[optind]) || (!amh_dev_set)) {
            usage();
            exit(22);
        }
    } else {
        if ((argc - optind) > 2) {
            usage();
            exit(22);
        }
	          	
        if (((argc - optind) < 2) || ((cmd = get_cmd(argv[optind + 1])) < 0)
	      || ((device = get_device(&argv[optind][1], cmd)) < 0)
	      || ((group = get_group(argv[optind])) < 0)
        )
        {
	    usage();
	    exit(22);
        }
    }
	
    /*
     * I thought about complaining here if repeats set and not a dimmer cmd,
     *   but it might be a good idea in some cases for reliability...
     *
     * It's also fun with -i.
     */
	
    unit = (unsigned char)((group << 4) | device);

    /*
     * Snag the serial port...
     */

    if ((fd = open(x10_dev, O_RDWR | O_NDELAY)) < 0) {
	fprintf(stderr, "%s:  Error %d opening %s.\n", MyName, errno,
	  x10_dev);
	perror(MyName);
	exit(errno);
    }

#ifdef DEBUG
    fprintf(stderr, "%s:  Opened %s.\n", MyName, x10_dev);
    fflush(stdout);
#endif

    /*
     * Do the deed...
     */

    while (repeats--) {
        x10_br_out(fd, (unsigned short)unit, (unsigned char)cmd);
        if (inverse_set) {
            if (cmd == ON) {
                cmd = OFF;
            } else {
                cmd = ON;
            }
        }
    }

    exit(0);
}
